import os


def service_providers_for_env():
    return os.getenv("PYCOIN_SERVICE_PROVIDERS", '').split(":")


def main_cache_dir():
    p = os.getenv("PYCOIN_CACHE_DIR")
    if p:
        p = os.path.expanduser(p)
    return p


def tx_read_cache_dirs():
    return [p for p in os.getenv("PYCOIN_TX_DB_DIRS", "").split(":") if len(p) > 0]


def tx_writable_cache_dir():
    p = main_cache_dir()
    if p:
        p = os.path.join(main_cache_dir(), "txs")
    return p



# -*- coding: utf-8 -*-
"""
A BIP0032-style hierarchical wallet.

Implement a BIP0032-style hierarchical wallet which can create public
or private wallet keys. Each key can create many child nodes. Each node
has a wallet key and a corresponding private & public key, which can
be used to generate Bitcoin addresses or WIF private keys.

At any stage, the private information can be stripped away, after which
descendants can only produce public keys.

Private keys can also generate "hardened" children, which cannot be
generated by the corresponding public keys. This is useful for generating
"change" addresses, for example, which there is no need to share with people
you give public keys to.


The MIT License (MIT)

Copyright (c) 2013 by Richard Kiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import hashlib
import hmac
import itertools
import struct

from .. import ecdsa

from ..encoding import public_pair_to_sec, sec_to_public_pair,\
    secret_exponent_to_wif, public_pair_to_bitcoin_address,\
    from_bytes_32, to_bytes_32,\
    public_pair_to_hash160_sec, EncodingError

from ..encoding import a2b_hashed_base58, b2a_hashed_base58
from ..networks import address_prefix_for_netcode, wif_prefix_for_netcode,\
    prv32_prefix_for_netcode, pub32_prefix_for_netcode
from .validate import netcode_and_type_for_data


class PublicPrivateMismatchError(Exception):
    pass


class InvalidKeyGeneratedError(Exception):
    pass


def wallet_iterator_for_wallet_key_path(wallet_key_path):
    subkey_paths = ""
    if "/" in wallet_key_path:
        wallet_key_path, subkey_paths = wallet_key_path.split("/", 1)
    return Wallet.from_wallet_key(wallet_key_path).subkeys_for_path(subkey_paths)


class Wallet(object):
    """
    This is a deterministic wallet that complies with BIP0032
    https://en.bitcoin.it/wiki/BIP_0032
    """
    @classmethod
    def from_master_secret(class_, master_secret, netcode='BTC'):
        """Generate a Wallet from a master password."""
        I64 = hmac.HMAC(key=b"Bitcoin seed", msg=master_secret, digestmod=hashlib.sha512).digest()
        return class_(
            is_private=True, netcode=netcode,
            chain_code=I64[32:], secret_exponent_bytes=I64[:32])

    @classmethod
    def from_wallet_key(class_, b58_str, allow_subkey_suffix=True):
        """Generate a Wallet from a base58 string in a standard way."""
        # TODO: support subkey suffixes

        data = a2b_hashed_base58(b58_str)
        netcode, key_type = netcode_and_type_for_data(data)

        if key_type not in ("pub32", "prv32"):
            raise EncodingError("bad wallet key header")

        is_private = (key_type == 'prv32')
        parent_fingerprint, child_number = struct.unpack(">4sL", data[5:13])

        d = dict(is_private=is_private, netcode=netcode, chain_code=data[13:45],
                 depth=ord(data[4:5]), parent_fingerprint=parent_fingerprint, child_number=child_number)

        if is_private:
            if data[45:46] != b'\0':
                raise EncodingError("private key encoded wrong")
            d["secret_exponent_bytes"] = data[46:]
        else:
            d["public_pair"] = sec_to_public_pair(data[45:])

        return class_(**d)

    def __init__(self, is_private, netcode, chain_code, depth=0, parent_fingerprint=b'\0\0\0\0',
                 child_number=0, secret_exponent_bytes=None, public_pair=None):
        """Don't use this. Use a classmethod to generate from a string instead."""
        if is_private:
            if public_pair:
                raise PublicPrivateMismatchError("can't include public_pair for private key")
        elif secret_exponent_bytes:
            raise PublicPrivateMismatchError("can't include secret_exponent_bytes for public key")
        self.is_private = is_private
        self.netcode = netcode
        if is_private:
            if len(secret_exponent_bytes) != 32:
                raise EncodingError("private key encoding wrong length")
            self.secret_exponent_bytes = secret_exponent_bytes
            self.secret_exponent = from_bytes_32(self.secret_exponent_bytes)
            if self.secret_exponent > ecdsa.generator_secp256k1.order():
                raise InvalidKeyGeneratedError(
                    "this key would produce an invalid secret exponent; please skip it")
            self.public_pair = ecdsa.public_pair_for_secret_exponent(ecdsa.generator_secp256k1,
                                                                     self.secret_exponent)
        else:
            self.public_pair = public_pair
        # validate public_pair is on the curve
        if not ecdsa.is_public_pair_valid(ecdsa.generator_secp256k1, self.public_pair):
            raise InvalidKeyGeneratedError("this key would produce an invalid public pair; please skip it")
        if not isinstance(chain_code, bytes):
            raise ValueError("chain code must be bytes")
        if len(chain_code) != 32:
            raise EncodingError("chain code wrong length")
        self.chain_code = chain_code
        self.depth = depth
        if len(parent_fingerprint) != 4:
            raise EncodingError("parent_fingerprint wrong length")
        self.parent_fingerprint = parent_fingerprint
        self.child_number = child_number
        self.subkey_cache = dict()

    def serialize(self, as_private=None):
        """Yield a 78-byte binary blob corresponding to this node."""
        if as_private is None:
            as_private = self.is_private
        if not self.is_private and as_private:
            raise PublicPrivateMismatchError("public key has no private parts")

        ba = bytearray()
        if as_private:
            ba.extend(prv32_prefix_for_netcode(self.netcode))
        else:
            ba.extend(pub32_prefix_for_netcode(self.netcode))
        ba.extend([self.depth])
        ba.extend(self.parent_fingerprint + struct.pack(">L", self.child_number) + self.chain_code)
        if as_private:
            ba += b'\0' + self.secret_exponent_bytes
        else:
            ba += public_pair_to_sec(self.public_pair, compressed=True)
        return bytes(ba)

    def fingerprint(self):
        return public_pair_to_hash160_sec(self.public_pair, compressed=True)[:4]

    def wallet_key(self, as_private=False):
        """Yield a 111-byte string corresponding to this node."""
        return b2a_hashed_base58(self.serialize(as_private=as_private))

    def wif(self, compressed=True):
        """Yield the WIF corresponding to this node."""
        if not self.is_private:
            raise PublicPrivateMismatchError("can't generate WIF for public key")
        return secret_exponent_to_wif(self.secret_exponent, compressed=compressed,
                                      wif_prefix=wif_prefix_for_netcode(self.netcode))

    def bitcoin_address(self, compressed=True):
        """Yield the Bitcoin address corresponding to this node."""
        return public_pair_to_bitcoin_address(self.public_pair, compressed=compressed,
                                              address_prefix=address_prefix_for_netcode(self.netcode))

    def public_copy(self):
        """Yield the corresponding public node for this node."""
        return self.__class__(is_private=False, netcode=self.netcode, chain_code=self.chain_code,
                              depth=self.depth, parent_fingerprint=self.parent_fingerprint,
                              child_number=self.child_number, public_pair=self.public_pair)

    def _subkey(self, i, is_hardened, as_private):
        """Yield a child node for this node.

        i: the index for this node.
        is_hardened: use "hardened key derivation". That is, the public version
            of this node cannot calculate this child.
        as_private: set to True to get a private subkey.

        Note that setting i<0 uses private key derivation, no matter the
        value for is_hardened."""
        if i > 0xffffffff:
            raise ValueError("i is too big: %d" % i)
        if i < 0:
            is_hardened = True
            i_as_bytes = struct.pack(">l", i)
        else:
            if i >= 0x80000000:
                raise ValueError("subkey index 0x%x too large" % i)
            i &= 0x7fffffff
            if is_hardened:
                i |= 0x80000000
            i_as_bytes = struct.pack(">L", i)
        if is_hardened:
            if not self.is_private:
                raise PublicPrivateMismatchError("can't derive a private key from a public key")
            data = b'\0' + self.secret_exponent_bytes + i_as_bytes
        else:
            data = public_pair_to_sec(self.public_pair, compressed=True) + i_as_bytes
        I64 = hmac.HMAC(key=self.chain_code, msg=data, digestmod=hashlib.sha512).digest()
        I_left_as_exponent = from_bytes_32(I64[:32])
        d = dict(is_private=as_private, netcode=self.netcode, chain_code=I64[32:],
                 depth=self.depth+1, parent_fingerprint=self.fingerprint(), child_number=i)

        if as_private:
            exponent = (I_left_as_exponent + self.secret_exponent) % ecdsa.generator_secp256k1.order()
            d["secret_exponent_bytes"] = to_bytes_32(exponent)
        else:
            x, y = self.public_pair
            the_point = I_left_as_exponent * ecdsa.generator_secp256k1 +\
                ecdsa.Point(ecdsa.generator_secp256k1.curve(), x, y, ecdsa.generator_secp256k1.order())
            d["public_pair"] = the_point.pair()
        return self.__class__(**d)

    def subkey(self, i=0, is_hardened=False, as_private=None):
        if as_private is None:
            as_private = self.is_private
        is_hardened = not not is_hardened
        as_private = not not as_private
        lookup = (i, is_hardened, as_private)
        if lookup not in self.subkey_cache:
            self.subkey_cache[lookup] = self._subkey(i, is_hardened, as_private)
        return self.subkey_cache[lookup]

    def subkey_for_path(self, path):
        """
        path: a path of subkeys denoted by numbers and slashes. Use
            H or i<0 for private key derivation. End with .pub to force
            the key public.

        Examples:
            1H/-5/2/1 would call subkey(i=1, is_hardened=True).subkey(i=-5).
                subkey(i=2).subkey(i=1) and then yield the private key
            0/0/458.pub would call subkey(i=0).subkey(i=0).subkey(i=458) and
                then yield the public key

        You should choose one of the p or the negative number convention for private key
        derivation and stick with it.
        """
        force_public = (path[-4:] == '.pub')
        if force_public:
            path = path[:-4]
        key = self
        if path:
            invocations = path.split("/")
            for v in invocations:
                is_hardened = v[-1] in ("'pH")
                if is_hardened:
                    v = v[:-1]
                v = int(v)
                key = key.subkey(i=v, is_hardened=is_hardened, as_private=key.is_private)
        if force_public and key.is_private:
            key = key.public_copy()
        return key

    def subkeys_for_path(self, path):
        """
        A generalized form that can return multiple subkeys.
        """
        if path == '':
            yield self
            return

        def range_iterator(the_range):
            for r in the_range.split(","):
                is_hardened = r[-1] in "'pH"
                if is_hardened:
                    r = r[:-1]
                hardened_char = "H" if is_hardened else ''
                if '-' in r:
                    low, high = [int(x) for x in r.split("-", 1)]
                    for t in range(low, high+1):
                        yield "%d%s" % (t, hardened_char)
                else:
                    yield "%s%s" % (r, hardened_char)

        def subkey_iterator(subkey_paths):
            # examples:
            #   0/1H/0-4 => ['0/1H/0', '0/1H/1', '0/1H/2', '0/1H/3', '0/1H/4']
            #   0/2,5,9-11 => ['0/2', '0/5', '0/9', '0/10', '0/11']
            #   3H/2/5/15-20p => ['3H/2/5/15p', '3H/2/5/16p', '3H/2/5/17p', '3H/2/5/18p',
            #          '3H/2/5/19p', '3H/2/5/20p']
            #   5-6/7-8p,15/1-2 => ['5/7H/1', '5/7H/2', '5/8H/1', '5/8H/2',
            #         '5/15/1', '5/15/2', '6/7H/1', '6/7H/2', '6/8H/1', '6/8H/2', '6/15/1', '6/15/2']

            components = subkey_paths.split("/")
            iterators = [range_iterator(c) for c in components]
            for v in itertools.product(*iterators):
                yield '/'.join(v)

        for subkey in subkey_iterator(path):
            yield self.subkey_for_path(subkey)

    def children(self, max_level=50, start_index=0, include_hardened=True):
        for i in range(start_index, max_level+start_index+1):
            yield self.subkey(i)
            if include_hardened:
                yield self.subkey(i, is_hardened=True)

    def __repr__(self):
        if self.child_number == 0:
            r = self.wallet_key(as_private=False)
        else:
            r = self.bitcoin_address()
        if self.is_private:
            return "private_for <%s>" % r
        return "<%s>" % r




from ... import ecdsa
from ... import encoding

from ...networks import DEFAULT_NETCODES, pay_to_script_prefix_for_netcode
from ...key.validate import netcode_and_type_for_data

from .ScriptPayToAddress import ScriptPayToAddress
from .ScriptPayToPublicKey import ScriptPayToPublicKey
from .ScriptPayToScript import ScriptPayToScript
from .ScriptMultisig import ScriptMultisig
from .ScriptUnknown import ScriptUnknown

SUBCLASSES = [ScriptPayToAddress, ScriptPayToPublicKey, ScriptPayToScript, ScriptMultisig, ScriptUnknown]


class SolvingError(Exception):
    pass


def script_obj_from_address(text, netcodes=DEFAULT_NETCODES):
    data = encoding.a2b_hashed_base58(text)
    netcode, key_type = netcode_and_type_for_data(data, netcodes=netcodes)
    if key_type == 'pay_to_script':
        return ScriptPayToScript(hash160=data[1:])
    if key_type == 'address':
        return ScriptPayToAddress(hash160=data[1:])
    raise ValueError("bad text")


def script_obj_from_script(script):
    for sc in SUBCLASSES:
        try:
            st = sc.from_script(script)
            return st
        except ValueError:
            pass
    return None


def address_for_pay_to_script(script, netcode="BTC"):
    address_prefix = pay_to_script_prefix_for_netcode(netcode)
    return encoding.hash160_sec_to_bitcoin_address(encoding.hash160(script), address_prefix=address_prefix)


def build_hash160_lookup(secret_exponents):
    d = {}
    for secret_exponent in secret_exponents:
        public_pair = ecdsa.public_pair_for_secret_exponent(ecdsa.generator_secp256k1, secret_exponent)
        for compressed in (True, False):
            hash160 = encoding.public_pair_to_hash160_sec(public_pair, compressed=compressed)
            d[hash160] = (secret_exponent, public_pair, compressed)
    return d


def build_p2sh_lookup(scripts):
    return dict((encoding.hash160(s), s) for s in scripts)




from .. import encoding
from ..networks import DEFAULT_NETCODES, NETWORK_NAMES, NETWORKS

DEFAULT_ADDRESS_TYPES = ["address", "pay_to_script"]

def _generate_network_prefixes():
    d = {}
    for n in NETWORKS:
        for prop in "wif address pay_to_script prv32 pub32".split():
            v = getattr(n, prop, None)
            if v:
                if v not in d:
                    d[v] = []
                d[v].append((n, prop))
    return d


NETWORK_PREFIXES = _generate_network_prefixes()


def netcode_and_type_for_data(data, netcodes=NETWORK_NAMES):
    """
    Given some already-decoded raw data from a base58 string,
    return (N, T) where N is the network code ("BTC" or "LTC") and
    T is the data type ("wif", "address", "public_pair", "prv32", "pub32").
    May also raise EncodingError.
    """
    d = {}
    for length in (4, 1):
        for network, the_type in NETWORK_PREFIXES.get(data[:length], []):
            d[network.code] = the_type
    for netcode in netcodes:
        v = d.get(netcode)
        if v:
            return netcode, v

    raise encoding.EncodingError("unknown prefix")


def _check_against(text, expected_type, allowable_netcodes):
    try:
        data = encoding.a2b_hashed_base58(text)
        netcode, the_type = netcode_and_type_for_data(data, netcodes=allowable_netcodes)
        if the_type in expected_type and netcode in allowable_netcodes:
            return netcode
    except encoding.EncodingError:
        pass
    return None


def is_address_valid(address, allowable_types=DEFAULT_ADDRESS_TYPES, allowable_netcodes=DEFAULT_NETCODES):
    """
    Accept an address, and a list of allowable address types (a subset of "address" and "pay_to_script"),
    and allowable networks (defaulting to just Bitcoin mainnet), return the network that the address is
    a part of, or None if it doesn't validate.
    """
    return _check_against(address, allowable_types, allowable_netcodes)


def is_wif_valid(wif, allowable_netcodes=DEFAULT_NETCODES):
    """
    Accept a WIF, and a list of allowable networks (defaulting to just Bitcoin mainnet), return
    the network that the wif is a part of, or None if it doesn't validate.
    """
    return _check_against(wif, ["wif"], allowable_netcodes)


def is_public_bip32_valid(hwif, allowable_netcodes=DEFAULT_NETCODES):
    """
    Accept a text representation of a BIP32 public wallet, and a list of allowable networks (defaulting
    to just Bitcoin mainnet), return the network that the wif is a part of, or None if it doesn't validate.
    """
    return _check_against(hwif, ["pub32"], allowable_netcodes)


def is_private_bip32_valid(hwif, allowable_netcodes=DEFAULT_NETCODES):
    """
    Accept a text representation of a BIP32 private wallet, and a list of allowable networks (defaulting
    to just Bitcoin mainnet), return the network that the wif is a part of, or None if it doesn't validate.
    """
    return _check_against(hwif, ["prv32"], allowable_netcodes)



from collections import namedtuple

from .serialize import h2b

NetworkValues = namedtuple('NetworkValues',
                           ('network_name', 'subnet_name', 'code', 'wif', 'address',
                            'pay_to_script', 'prv32', 'pub32'))

NETWORKS = (
    NetworkValues("Bitcoin", "mainnet", "BTC", b'\x80', b'\0', b'\5', h2b("0488ADE4"), h2b("0488B21E")),
    NetworkValues("Bitcoin", "testnet3", "XTN", b'\xef', b'\x6f', b'\xc4',
                  h2b("04358394"), h2b("043587CF")),
    NetworkValues("Litecoin", "mainnet", "LTC", b'\xb0', b'\x30', b'\5', h2b('019d9cfe'), h2b('019da462')),
    NetworkValues("Litecoin", "testnet", "XLT", b'\xef', b'\x6f', b'\xc4', h2b('0436ef7d'), h2b('0436f6e1')),
    NetworkValues("Dogecoin", "mainnet", "DOGE", b'\x9e', b'\x1e', b'\x16',
                  h2b("02fda4e8"), h2b("02fda923")),
    # BlackCoin: unsure about bip32 prefixes; assuming will use Bitcoin's
    NetworkValues("Blackcoin", "mainnet", "BLK", b'\x99', b'\x19', None, h2b("0488ADE4"), h2b("0488B21E")),
    NetworkValues("Viacoin", "mainnet", "VIA", b'\xc7', b'\x47', b'\x21', h2b('0488ADE4'), h2b('0488B21E')),
    NetworkValues("Viacoin", "testnet", "TVI", b'\xff', b'\x7f', b'\xc4', h2b('04358394'), h2b('043587CF')),
)


# Map from short code to details about that network.
NETWORK_NAME_LOOKUP = dict((i.code, i) for i in NETWORKS)

# All network names, return in same order as list above: for UI purposes.
NETWORK_NAMES = [i.code for i in NETWORKS]

DEFAULT_NETCODES = NETWORK_NAMES


def _lookup(netcode, property):
    # Lookup a specific value needed for a specific network
    network = NETWORK_NAME_LOOKUP.get(netcode)
    if network:
        return getattr(network, property)
    return None


def network_name_for_netcode(netcode):
    return _lookup(netcode, "network_name")


def subnet_name_for_netcode(netcode):
    return _lookup(netcode, "subnet_name")


def full_network_name_for_netcode(netcode):
    network = NETWORK_NAME_LOOKUP[netcode]
    if network:
        return "%s %s" % (network.network_name, network.subnet_name)


def wif_prefix_for_netcode(netcode):
    return _lookup(netcode, "wif")


def address_prefix_for_netcode(netcode):
    return _lookup(netcode, "address")


def pay_to_script_prefix_for_netcode(netcode):
    return _lookup(netcode, "pay_to_script")


def prv32_prefix_for_netcode(netcode):
    return _lookup(netcode, "prv32")


def pub32_prefix_for_netcode(netcode):
    return _lookup(netcode, "pub32")



from ..script import opcodes, tools
from ..script.vm import parse_signature_blob
from ..script.microcode import VCH_TRUE

from ... import ecdsa
from ... import encoding

from ...networks import address_prefix_for_netcode
from ...serialize import b2h

from .ScriptType import ScriptType


# see BIP11 https://github.com/bitcoin/bips/blob/master/bip-0011.mediawiki

class ScriptMultisig(ScriptType):

    def __init__(self, n, sec_keys):
        self.n = n
        self.sec_keys = sec_keys
        self._script = None

    @classmethod
    def from_script(cls, script):
        pc = 0
        if len(script) == 0:
            raise ValueError("blank script")
        opcode, data, pc = tools.get_opcode(script, pc)

        if not opcodes.OP_1 <= opcode < opcodes.OP_16:
            raise ValueError("n value invalid")
        n = opcode + (1 - opcodes.OP_1)
        sec_keys = []
        while 1:
            if pc >= len(script):
                raise ValueError("unexpected end of script")
            opcode, data, pc = tools.get_opcode(script, pc)
            l = len(data)
            if l < 33 or l > 120:
                break
            sec_keys.append(data)
        m = opcode + (1 - opcodes.OP_1)
        if n > m or len(sec_keys) != m:
            raise ValueError("m value wrong")

        opcode, data, pc = tools.get_opcode(script, pc)
        if opcode != opcodes.OP_CHECKMULTISIG:
            raise ValueError("no OP_CHECKMULTISIG")
        if pc != len(script):
            raise ValueError("extra stuff at end")

        return cls(sec_keys=sec_keys, n=n)

    def address(self):
        if self._address is None:
            self._address = encoding.hash160_sec_to_bitcoin_address(
                self.hash160, address_prefix=self.address_prefix)
        return self._address

    def script(self):
        if self._script is None:
            # create the script
            # TEMPLATE = m {pubkey}...{pubkey} n OP_CHECKMULTISIG
            public_keys = [b2h(sk) for sk in self.sec_keys]
            script_source = "%d %s %d OP_CHECKMULTISIG" % (self.n, " ".join(public_keys), len(public_keys))
            self._script = tools.compile(script_source)
        return self._script

    def solve(self, **kwargs):
        """
        The kwargs required depend upon the script type.
        hash160_lookup:
            dict-like structure that returns a secret exponent for a hash160
        existing_script:
            existing solution to improve upon (optional)
        sign_value:
            the integer value to sign (derived from the transaction hash)
        signature_type:
            usually SIGHASH_ALL (1)
        """
        # we need a hash160 => secret_exponent lookup
        db = kwargs.get("hash160_lookup")
        if db is None:
            raise SolvingError("missing hash160_lookup parameter")

        sign_value = kwargs.get("sign_value")
        signature_type = kwargs.get("signature_type")

        secs_solved = set()
        existing_signatures = []
        existing_script = kwargs.get("existing_script")
        if existing_script:
            pc = 0
            opcode, data, pc = tools.get_opcode(existing_script, pc)
            # ignore the first opcode
            while pc < len(existing_script):
                opcode, data, pc = tools.get_opcode(existing_script, pc)
                sig_pair, actual_signature_type = parse_signature_blob(data)
                for sec_key in self.sec_keys:
                    try:
                        public_pair = encoding.sec_to_public_pair(sec_key)
                        sig_pair, signature_type = parse_signature_blob(data)
                        v = ecdsa.verify(ecdsa.generator_secp256k1, public_pair, sign_value, sig_pair)
                        if v:
                            existing_signatures.append(data)
                            secs_solved.add(sec_key)
                            break
                    except encoding.EncodingError:
                        # if public_pair is invalid, we just ignore it
                        pass

        for sec_key in self.sec_keys:
            if sec_key in secs_solved:
                continue
            if len(existing_signatures) >= self.n:
                break
            hash160 = encoding.hash160(sec_key)
            result = db.get(hash160)
            if result is None:
                continue
            secret_exponent, public_pair, compressed = result
            binary_signature = self._create_script_signature(secret_exponent, sign_value, signature_type)
            existing_signatures.append(binary_signature)
        DUMMY_SIGNATURE = self._dummy_signature(signature_type)
        while len(existing_signatures) < self.n:
            existing_signatures.append(DUMMY_SIGNATURE)

        script = "OP_0 %s" % " ".join(b2h(s) for s in existing_signatures)
        solution = tools.compile(script)
        return solution

    def info(self, netcode='BTC'):
        address_prefix = address_prefix_for_netcode(netcode)
        hash160s = [encoding.hash160(sk) for sk in self.sec_keys]
        addresses = [encoding.hash160_sec_to_bitcoin_address(h1, address_prefix=address_prefix)
                     for h1 in hash160s]
        return dict(type="multisig m of n", m=len(self.sec_keys), n=self.n, addresses=addresses,
                    hash160s=hash160s, script=self._script, address_prefix=address_prefix,
                    summary="%d of %s" % (self.n, addresses))

    def __repr__(self):
        info = self.info()
        return "<Script: multisig %d of %d (%s)>" % (info["n"], info["m"], "/".join(info["addresses"]))



import json
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

from pycoin.convention import btc_to_satoshi
from pycoin.serialize import b2h_rev, h2b, h2b_rev
from pycoin.tx.Tx import Tx, TxIn, TxOut
from pycoin.tx.script import tools


def get_json_for_hash(the_hash):
    d = urlopen("http://blockexplorer.com/rawtx/%s" % b2h_rev(the_hash)).read()
    return json.loads(d.decode("utf8"))


def get_tx(tx_hash):
    """
    Get a Tx by its hash.
    """
    # TODO: fix this
    j = get_json_for_hash(tx_hash)
    txs_in = []
    for j_in in j.get("in"):
        if j_in.get("coinbase"):
            txs_in.append(TxIn.coinbase_tx_in(h2b(j_in["coinbase"])))
        else:
            txs_in.append(TxIn(
                h2b_rev(j_in["prev_out"]["hash"]),
                int(j_in["prev_out"]["n"]),
                tools.compile(j_in["scriptSig"])))

    txs_out = []
    for j_out in j.get("out"):
        txs_out.append(TxOut(int(btc_to_satoshi(j_out["value"])), tools.compile(j_out["scriptPubKey"])))

    tx = Tx(int(j["ver"]), txs_in, txs_out, int(j["lock_time"]))
    assert tx.hash() == tx_hash
    return tx



#!/usr/bin/env python

import argparse
import codecs
import io
import re

from pycoin.serialize import h2b_rev
from pycoin.services import get_tx_db
from pycoin.tx import Tx


def main():
    parser = argparse.ArgumentParser(description="Add a transaction to tx cache.")
    parser.add_argument("tx_id_or_path", nargs="+",
                        help='The id of the transaction to fetch from web services or the path to it.')

    args = parser.parse_args()

    TX_RE = re.compile(r"^[0-9a-fA-F]{64}$")

    tx_db = get_tx_db()

    for p in args.tx_id_or_path:
        if TX_RE.match(p):
            tx = tx_db.get(h2b_rev(p))
            if not tx:
                parser.error("can't find Tx with id %s" % p)
        else:
            f = open(p, "rb")
            try:
                if f.name.endswith("hex"):
                    f = io.BytesIO(codecs.getreader("hex_codec")(f).read())
                tx = Tx.parse(f)
            except Exception:
                parser.error("can't parse %s" % f.name)

        tx_db[tx.hash()] = tx
        print("cached %s" % tx.id())

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Sample usage (with fake coinbase transaction):
# ./spend.py `./spend.py -c KxwGdpvzjzD5r6Qwg5Ev7gAv2Wn53tmSFfingBThhJEThQFcWPdj/20` 19TKi9Mv8AVLguAVYyCTY5twy5PKoEqrRf/19.9999 -p KxwGdpvzjzD5r6Qwg5Ev7gAv2Wn53tmSFfingBThhJEThQFcWPdj

import argparse
import decimal
import io
import itertools
import sys

from pycoin import ecdsa
from pycoin import encoding
from pycoin.convention import tx_fee, btc_to_satoshi, satoshi_to_btc
from pycoin.serialize import b2h, h2b
from pycoin.services import blockchain_info
from pycoin.tx import Tx, UnsignedTx, TxOut, SecretExponentSolver
from pycoin.wallet import Wallet

def roundrobin(*iterables):
    "roundrobin('ABC', 'D', 'EF') --> A D E B F C"
    # Recipe credited to George Sakkis
    pending = len(iterables)
    nexts = itertools.cycle(iter(it).next for it in iterables)
    while pending:
        try:
            for next in nexts:
                yield next()
        except StopIteration:
            pending -= 1
            nexts = itertools.cycle(itertools.islice(nexts, pending))

def secret_exponents_iterator(wif_file, private_keys):
    def private_key_iterator(pk):
        try:
            wallet = Wallet.from_wallet_key(pk)
            return (w.secret_exponent for w in wallet.children(max_level=50, start_index=0))
        except (encoding.EncodingError, TypeError):
            try:
                exp = encoding.wif_to_secret_exponent(pk)
                return [exp]
            except encoding.EncodingError:
                sys.stderr.write('bad value: "%s"\n' % pk)
                sys.exit(1)

    iterables = []
    if wif_file:
        for l in wif_file:
            iterables.append(private_key_iterator(l[:-1]))
    if private_keys:
        for pk in private_keys:
            iterables.append(private_key_iterator(pk))
    for v in roundrobin(*iterables):
        yield v

def calculate_fees(unsigned_tx):
    total_value = sum(unsigned_tx_out.coin_value for unsigned_tx_out in unsigned_tx.unsigned_txs_out)
    total_spent = sum(tx_out.coin_value for tx_out in unsigned_tx.new_txs_out)
    return total_value, total_spent

def check_fees(unsigned_tx):
    total_value, total_spent = calculate_fees(unsigned_tx)
    actual_tx_fee = total_value - total_spent
    recommended_tx_fee = tx_fee.recommended_fee_for_tx(unsigned_tx)
    if actual_tx_fee > recommended_tx_fee:
        print("warning: transaction fee of exceeds expected value of %s BTC" % satoshi_to_btc(recommended_tx_fee))
    elif actual_tx_fee < 0:
        print("not enough source coins (%s BTC) for destination (%s BTC). Short %s BTC" % (satoshi_to_btc(total_value), satoshi_to_btc(total_spent), satoshi_to_btc(-actual_tx_fee)))
    elif actual_tx_fee < recommended_tx_fee:
        print("warning: transaction fee lower than (casually calculated) expected value of %s BTC, transaction might not propogate" % satoshi_to_btc(recommended_tx_fee))
    return actual_tx_fee

def get_unsigned_tx(parser):
    args = parser.parse_args()
    # if there is only one item passed, it's assumed to be hex
    if len(args.txinfo) == 1:
        try:
            s = io.BytesIO(h2b(args.txinfo[0]))
            return UnsignedTx.parse(s)
        except Exception:
            parser.error("can't parse %s as hex\n" % args.txinfo[0])

    coins_from = []
    coins_to = []
    for txinfo in args.txinfo:
        if '/' in txinfo:
            parts = txinfo.split("/")
            if len(parts) == 2:
                # we assume it's an output
                address, amount = parts
                amount = btc_to_satoshi(amount)
                coins_to.append((amount, address))
            else:
                try:
                    # we assume it's an input of the form
                    #  tx_hash_hex/tx_output_index_decimal/tx_out_val/tx_out_script_hex
                    tx_hash_hex, tx_output_index_decimal, tx_out_val, tx_out_script_hex = parts
                    tx_hash = h2b(tx_hash_hex)
                    tx_output_index = int(tx_output_index_decimal)
                    tx_out_val = btc_to_satoshi(decimal.Decimal(tx_out_val))
                    tx_out_script = h2b(tx_out_script_hex)
                    tx_out = TxOut(tx_out_val, tx_out_script)
                    coins_source = (tx_hash, tx_output_index, tx_out)
                    coins_from.append(coins_source)
                except Exception:
                    parser.error("can't parse %s\n" % txinfo)
        else:
            print("looking up funds for %s from blockchain.info" % txinfo)
            coins_sources = blockchain_info.unspent_tx_outs_info_for_address(txinfo)
            coins_from.extend(coins_sources)

    unsigned_tx = UnsignedTx.standard_tx(coins_from, coins_to)
    return unsigned_tx

def create_coinbase_tx(parser):
    args = parser.parse_args()
    try:
        if len(args.txinfo) != 1:
            parser.error("coinbase transactions need exactly one output parameter (wif/BTC count)")
        wif, btc_amount = args.txinfo[0].split("/")
        satoshi_amount = btc_to_satoshi(btc_amount)
        secret_exponent, compressed = encoding.wif_to_tuple_of_secret_exponent_compressed(wif)
        public_pair = ecdsa.public_pair_for_secret_exponent(ecdsa.secp256k1.generator_secp256k1, secret_exponent)
        public_key_sec = encoding.public_pair_to_sec(public_pair, compressed=compressed)
        coinbase_tx = Tx.coinbase_tx(public_key_sec, satoshi_amount)
        return coinbase_tx
    except Exception:
        parser.error("coinbase transactions need exactly one output parameter (wif/BTC count)")

EPILOG = "If you generate an unsigned transaction, the output is a hex dump that can be used by this script on an air-gapped machine."

def main():
    parser = argparse.ArgumentParser(description="Create a Bitcoin transaction.", epilog=EPILOG)

    parser.add_argument('-g', "--generate-unsigned", help='generate unsigned transaction', action='store_true')
    parser.add_argument('-f', "--private-key-file", help='file containing WIF or BIP0032 private keys', metavar="path-to-file-with-private-keys", type=argparse.FileType('r'))
    parser.add_argument('-p', "--private-key", help='WIF or BIP0032 private key', metavar="private-key", type=str, nargs="+")
    parser.add_argument('-c', "--coinbase", help='Create a (bogus) coinbase transaction. For testing purposes. You must include exactly one WIF in this case.', action='store_true')
    parser.add_argument("txinfo", help='either a hex dump of the unsigned transaction, or a list of bitcoin addresses with optional "/value" if they are destination addresses', nargs="+")

    args = parser.parse_args()
    if args.coinbase:
        new_tx = create_coinbase_tx(parser)
        tx_hash_hex = b2h(new_tx.hash())
        tx_output_index = 0
        tx_out_val = str(satoshi_to_btc(new_tx.txs_out[tx_output_index].coin_value))
        tx_out_script_hex = b2h(new_tx.txs_out[tx_output_index].script)
        # product output in the form:
        #  tx_hash_hex/tx_output_index_decimal/tx_out_val/tx_out_script_hex
        # which can be used as a fake input to a later transaction
        print("/".join([tx_hash_hex, str(tx_output_index), tx_out_val, tx_out_script_hex]))
        return

    unsigned_tx = get_unsigned_tx(parser)
    actual_tx_fee = check_fees(unsigned_tx)
    if actual_tx_fee < 0:
        sys.exit(1)
    print("transaction fee: %s BTC" % satoshi_to_btc(actual_tx_fee))

    if args.generate_unsigned:
        s = io.BytesIO()
        unsigned_tx.stream(s)
        tx_bytes = s.getvalue()
        tx_hex = b2h(tx_bytes)
        print(tx_hex)
        sys.exit(0)

    secret_exponents = secret_exponents_iterator(args.private_key_file, args.private_key)
    solver = SecretExponentSolver(secret_exponents)
    new_tx = unsigned_tx.sign(solver)
    s = io.BytesIO()
    new_tx.stream(s)
    tx_bytes = s.getvalue()
    tx_hex = b2h(tx_bytes)
    print("copy the following hex to http://blockchain.info/pushtx to put the transaction on the network:\n")
    print(tx_hex)

if __name__ == '__main__':
    main()



from pycoin.serialize import b2h, b2h_rev

try:
    from bitcoinrpc.authproxy import AuthServiceProxy
except ImportError:
    print("This script depends upon python-bitcoinrpc.")
    print("pip install -e git+https://github.com/jgarzik/python-bitcoinrpc#egg=python_bitcoinrpc-master")
    raise


def unspent_to_bitcoind_dict(tx_in, tx_out):
    return dict(
        txid=b2h_rev(tx_in.previous_hash),
        vout=tx_in.previous_index,
        scriptPubKey=b2h(tx_out.script)
    )


def bitcoind_agrees_on_transaction_validity(bitcoind_url, tx):
    connection = AuthServiceProxy(bitcoind_url)
    tx.check_unspents()
    unknown_tx_outs = [unspent_to_bitcoind_dict(tx_in, tx_out)
                       for tx_in, tx_out in zip(tx.txs_in, tx.unspents)]
    signed = connection.signrawtransaction(tx.as_hex(), unknown_tx_outs, [])
    is_ok = [tx.is_signature_ok(idx) for idx in range(len(tx.txs_in))]
    return all(is_ok) == signed.get("complete")



from .ScriptType import ScriptType


class ScriptUnknown(ScriptType):
    def __init__(self, script):
        self._script = script

    @classmethod
    def from_script(cls, script):
        return cls(script)

    def script(self):
        return self._script

    def solve(self, **kwargs):
        raise SolvingError("unknown script type")

    def info(self, netcode='BTC'):
        address = "(unknown)"
        return dict(type="unknown script", address=address, script=self._script, summary=address)

    def __repr__(self):
        return "<Script: unknown of size %d>" % len(self._script)



import json

try:
    from urllib2 import urlopen, Request
except ImportError:
    from urllib.request import urlopen, Request

from pycoin.serialize import h2b_rev
from pycoin.tx import Spendable
from pycoin.tx.script import tools


def spendables_for_address(bitcoin_address):
    """
    Return a list of Spendable objects for the
    given bitcoin address.
    """
    URL = "https://api.biteasy.com/blockchain/v1/addresses/%s/unspent-outputs" % bitcoin_address
    r = Request(URL,
                headers={"content-type": "application/json", "accept": "*/*", "User-Agent": "curl/7.29.0"})
    d = urlopen(r).read()
    json_response = json.loads(d.decode("utf8"))
    spendables = []
    for tx_out_info in json_response.get("data", {}).get("outputs"):
        if tx_out_info.get("to_address") == bitcoin_address:
            coin_value = tx_out_info["value"]
            script = tools.compile(tx_out_info.get("script_pub_key"))
            previous_hash = h2b_rev(tx_out_info.get("transaction_hash"))
            previous_index = tx_out_info.get("transaction_index")
            spendables.append(Spendable(coin_value, script, previous_hash, previous_index))
    return spendables



import collections

from pycoin import ecdsa

from ..script import der, opcodes, tools


bytes_from_int = chr if bytes == str else lambda x: bytes([x])


class ScriptType(object):
    def __init__(self):
        raise NotImplemented()

    @classmethod
    def subclasses(cls, skip_self=True):
        for c in cls.__subclasses__():
            for c1 in c.subclasses(skip_self=False):
                yield c1
            if not skip_self:
                yield cls

    @classmethod
    def from_address(cls, text, netcodes=None):
        for sc in cls.subclasses():
            try:
                st = sc.from_address(text, netcodes=netcodes)
                return st
            except Exception:
                pass

    @classmethod
    def from_script(cls, script, netcode="BTC"):
        for sc in cls.subclasses():
            try:
                st = sc.from_script(script)
                return st
            except Exception:
                pass

    @classmethod
    def match(cls, script):
        template = cls.TEMPLATE
        r = collections.defaultdict(list)
        pc1 = pc2 = 0
        while 1:
            if pc1 == len(script) and pc2 == len(template):
                return r
            if pc1 >= len(script) or pc2 >= len(template):
                break
            opcode1, data1, pc1 = tools.get_opcode(script, pc1)
            opcode2, data2, pc2 = tools.get_opcode(template, pc2)
            if opcode2 == opcodes.OP_PUBKEY:
                l1 = len(data1)
                if l1 < 33 or l1 > 120:
                    break
                r["PUBKEY_LIST"].append(data1)
            elif opcode2 == opcodes.OP_PUBKEYHASH:
                if len(data1) != 160/8:
                    break
                r["PUBKEYHASH_LIST"].append(data1)
            elif (opcode1, data1) != (opcode2, data2):
                break
        raise ValueError("script doesn't match")

    def _create_script_signature(self, secret_exponent, sign_value, signature_type):
        order = ecdsa.generator_secp256k1.order()
        r, s = ecdsa.sign(ecdsa.generator_secp256k1, secret_exponent, sign_value)
        if s + s > order:
            s = order - s
        return der.sigencode_der(r, s) + bytes_from_int(signature_type)

    def _dummy_signature(self, signature_type):
        order = ecdsa.generator_secp256k1.order()
        r, s = order - 1, order // 2
        return der.sigencode_der(r, s) + bytes_from_int(signature_type)

    def address(self, netcode='BTC'):
        return self.info()["address"]

    def solve(self, **kwargs):
        """
        The kwargs required depend upon the script type.
        """
        raise NotImplemented()



from pycoin import ecdsa
from pycoin.key.validate import netcode_and_type_for_data
from pycoin.networks import address_prefix_for_netcode, wif_prefix_for_netcode

from pycoin.encoding import a2b_hashed_base58, secret_exponent_to_wif,\
    public_pair_to_sec, hash160,\
    hash160_sec_to_bitcoin_address, sec_to_public_pair,\
    is_sec_compressed, from_bytes_32, EncodingError

from .bip32 import Wallet


class Key(object):
    def __init__(self, hierarchical_wallet=None, secret_exponent=None, public_pair=None, hash160=None,
                 prefer_uncompressed=None, is_compressed=True, netcode='BTC'):
        """
        hierarchical_wallet:
            a bip32 wallet
        secret_exponent:
            a long representing the secret exponent
        public_pair:
            a tuple of long integers on the ecdsa curve
        hash160:
            a hash160 value corresponding to a bitcoin address
        Include at most one of hierarchical_wallet, secret_exponent, public_pair or hash160.
        prefer_uncompressed, is_compressed (booleans) are optional.
        netcode:
            the code for the network (as defined in pycoin.networks)
        """
        if [hierarchical_wallet, secret_exponent, public_pair, hash160].count(None) != 3:
            raise ValueError("exactly one of hierarchical_wallet, secret_exponent, public_pair, hash160"
                             " must be passed.")
        if prefer_uncompressed is None:
            prefer_uncompressed = not is_compressed
        self._prefer_uncompressed = prefer_uncompressed
        self._hierarchical_wallet = hierarchical_wallet
        self._secret_exponent = secret_exponent
        self._public_pair = public_pair
        if hash160:
            if is_compressed:
                self._hash160_compressed = hash160
            else:
                self._hash160_uncompressed = hash160
        self._netcode = netcode
        self._calculate_all()

    @classmethod
    def from_text(class_, text, is_compressed=True):
        """
        This function will accept a BIP0032 wallet string, a WIF, or a bitcoin address.

        The "is_compressed" parameter is ignored unless a public address is passed in.
        """

        data = a2b_hashed_base58(text)
        netcode, key_type = netcode_and_type_for_data(data)
        data = data[1:]

        if key_type in ("pub32", "prv32"):
            hw = Wallet.from_wallet_key(text)
            return Key(hierarchical_wallet=hw, netcode=netcode)

        if key_type == 'wif':
            is_compressed = (len(data) > 32)
            if is_compressed:
                data = data[:-1]
            return Key(
                secret_exponent=from_bytes_32(data),
                prefer_uncompressed=not is_compressed, netcode=netcode)
        if key_type == 'address':
            return Key(hash160=data, is_compressed=is_compressed, netcode=netcode)
        raise EncodingError("unknown text: %s" % text)

    @classmethod
    def from_sec(class_, sec):
        """
        Create a key from an sec bytestream (which is an encoding of a public pair).
        """
        public_pair = sec_to_public_pair(sec)
        return Key(public_pair=public_pair, prefer_uncompressed=not is_sec_compressed(sec))

    def public_copy(self):
        """
        Create a copy of this key with private key information removed.
        """
        if self._hierarchical_wallet:
            return Key(hierarchical_wallet=self._hierarchical_wallet.public_copy())
        if self.public_pair():
            return Key(public_pair=self.public_pair())
        return self

    def _calculate_all(self):
        for attr in "_secret_exponent _public_pair _wif_uncompressed _wif_compressed _sec_compressed" \
                " _sec_uncompressed _hash160_compressed _hash160_uncompressed _address_compressed" \
                " _address_uncompressed _netcode".split():
                setattr(self, attr, getattr(self, attr, None))

        if self._hierarchical_wallet:
            if self._hierarchical_wallet.is_private:
                self._secret_exponent = self._hierarchical_wallet.secret_exponent
            else:
                self._public_pair = self._hierarchical_wallet.public_pair
            self._netcode = self._hierarchical_wallet.netcode

        wif_prefix = wif_prefix_for_netcode(self._netcode)

        if self._secret_exponent:
            self._wif_uncompressed = secret_exponent_to_wif(
                self._secret_exponent, compressed=False, wif_prefix=wif_prefix)
            self._wif_compressed = secret_exponent_to_wif(
                self._secret_exponent, compressed=True, wif_prefix=wif_prefix)
            self._public_pair = ecdsa.public_pair_for_secret_exponent(
                ecdsa.generator_secp256k1, self._secret_exponent)

        if self._public_pair:
            self._sec_compressed = public_pair_to_sec(self._public_pair, compressed=True)
            self._sec_uncompressed = public_pair_to_sec(self._public_pair, compressed=False)
            self._hash160_compressed = hash160(self._sec_compressed)
            self._hash160_uncompressed = hash160(self._sec_uncompressed)

        address_prefix = address_prefix_for_netcode(self._netcode)

        if self._hash160_compressed:
            self._address_compressed = hash160_sec_to_bitcoin_address(
                self._hash160_compressed, address_prefix=address_prefix)

        if self._hash160_uncompressed:
            self._address_uncompressed = hash160_sec_to_bitcoin_address(
                self._hash160_uncompressed, address_prefix=address_prefix)

    def as_text(self):
        """
        Return a textual representation of this key.
        """
        if self._hierarchical_wallet:
            return self._hierarchical_wallet.wallet_key(as_private=self._hierarchical_wallet.is_private)
        if self._secret_exponent:
            return self.wif()
        return self.address()

    def hierarchical_wallet(self):
        return self._hierarchical_wallet

    def hwif(self, as_private=False):
        """
        Return a textual representation of the hiearachical wallet (reduced to public), or None.
        """
        if self._hierarchical_wallet:
            return self._hierarchical_wallet.wallet_key(as_private=as_private)
        return None

    def secret_exponent(self):
        """
        Return an integer representing the secret exponent (or None).
        """
        return self._secret_exponent

    def public_pair(self):
        """
        Return a pair of integers representing the public key (or None).
        """
        return self._public_pair

    def _use_uncompressed(self, use_uncompressed):
        if use_uncompressed:
            return use_uncompressed
        if use_uncompressed is None:
            return self._prefer_uncompressed
        return False

    def wif(self, use_uncompressed=None):
        """
        Return the WIF representation of this key, if available.
        If use_uncompressed is not set, the preferred representation is returned.
        """
        if self._use_uncompressed(use_uncompressed):
            return self._wif_uncompressed
        return self._wif_compressed

    def sec(self, use_uncompressed=None):
        """
        Return the SEC representation of this key, if available.
        If use_uncompressed is not set, the preferred representation is returned.
        """
        if self._use_uncompressed(use_uncompressed):
            return self._sec_uncompressed
        return self._sec_compressed

    def hash160(self, use_uncompressed=None):
        """
        Return the hash160 representation of this key, if available.
        If use_uncompressed is not set, the preferred representation is returned.
        """
        if self._use_uncompressed(use_uncompressed):
            return self._hash160_uncompressed
        return self._hash160_compressed

    def address(self, use_uncompressed=None):
        """
        Return the public address representation of this key, if available.
        If use_uncompressed is not set, the preferred representation is returned.
        """
        if self._use_uncompressed(use_uncompressed):
            return self._address_uncompressed
        return self._address_compressed

    def subkey(self, path_to_subkey):
        """
        Return the Key corresponding to the hierarchical wallet's subkey (or None).
        """
        if self._hierarchical_wallet:
            return Key(hierarchical_wallet=self._hierarchical_wallet.subkey_for_path(str(path_to_subkey)))

    def subkeys(self, path_to_subkeys):
        """
        Return an iterator yielding Keys corresponding to the
        hierarchical wallet's subkey path (or just this key).
        """
        if self._hierarchical_wallet:
            for subwallet in self._hierarchical_wallet.subkeys_for_path(path_to_subkeys):
                yield Key(hierarchical_wallet=subwallet)
        else:
            yield self



from ..convention import satoshi_to_mbtc
from ..serialize import b2h, b2h_rev, h2b, h2b_rev
from ..serialize.bitcoin_streamer import parse_struct, stream_struct

from .TxIn import TxIn
from .TxOut import TxOut


class Spendable(TxOut):
    def __init__(self, coin_value, script, tx_hash, tx_out_index, block_index_available=0,
                 does_seem_spent=False, block_index_spent=0):
        self.coin_value = int(coin_value)
        self.script = script
        self.tx_hash = tx_hash
        self.tx_out_index = tx_out_index
        self.block_index_available = block_index_available
        self.does_seem_spent = bool(does_seem_spent)
        self.block_index_spent = block_index_spent

    def stream(self, f, as_spendable=False):
        super(Spendable, self).stream(f)
        if as_spendable:
            stream_struct("#LIbI", f, self.previous_hash, self.previous_index,
                          self.block_index_available, self.does_seem_spent, self.block_index_spent)

    @classmethod
    def parse(class_, f):
        return class_(*parse_struct("QS#LIbI", f))

    def as_dict(self):
        # for use with JSON
        return dict(
            coin_value=self.coin_value,
            script_hex=b2h(self.script),
            tx_hash_hex=b2h(self.tx_hash),
            tx_out_index=self.tx_out_index,
            block_index_available=self.block_index_available,
            does_seem_spent=int(self.does_seem_spent),
            block_index_spent=self.block_index_spent
        )

    @classmethod
    def from_dict(class_, d):
        return class_(d["coin_value"], h2b(d["script_hex"]),
                      h2b(d["tx_hash_hex"]), d["tx_out_index"],
                      d.get("block_index_available", 0), d.get("does_seem_spent", 0),
                      d.get("block_index_spent", 0))

    def as_text(self):
        return "/".join([b2h_rev(self.tx_hash), str(self.tx_out_index), b2h(self.script),
                         str(self.coin_value), str(self.block_index_available),
                         "%d" % self.does_seem_spent, str(self.block_index_spent)])

    @classmethod
    def from_text(class_, text):
        the_tuple = (text.split("/") + [0, 0, 0])[:7]
        tx_hash_hex, tx_out_index_str, script_hex, coin_value, \
            block_index_available, does_seem_spent, block_index_spent = the_tuple
        tx_hash = h2b_rev(tx_hash_hex)
        tx_out_index = int(tx_out_index_str)
        script = h2b(script_hex)
        coin_value = int(coin_value)
        return class_(coin_value, script, tx_hash, tx_out_index, block_index_available,
                      bool(does_seem_spent), block_index_spent)

    def tx_in(self, script=b'', sequence=4294967295):
        return TxIn(self.tx_hash, self.tx_out_index, script, sequence)

    def __str__(self):
        return 'Spendable<%s mbtc "%s:%d">' % (
            satoshi_to_mbtc(self.coin_value), b2h_rev(self.tx_hash), self.tx_out_index)

    def __repr__(self):
        return str(self)




import struct

from .streamer import Streamer


def parse_bc_int(f):
    v = ord(f.read(1))
    if v == 253:
        v = struct.unpack("<H", f.read(2))[0]
    elif v == 254:
        v = struct.unpack("<L", f.read(4))[0]
    elif v == 255:
        v = struct.unpack("<Q", f.read(8))[0]
    return v


def parse_bc_string(f):
    size = parse_bc_int(f)
    return f.read(size)


def stream_bc_int(f, v):
    if v < 253:
        f.write(struct.pack("<B", v))
    elif v <= 65535:
        f.write(b'\xfd' + struct.pack("<H", v))
    elif v <= 0xffffffff:
        f.write(b'\xfe' + struct.pack("<L", v))
    else:
        f.write(b'\xff' + struct.pack("<Q", v))


def stream_bc_string(f, v):
    stream_bc_int(f, len(v))
    f.write(v)

STREAMER_FUNCTIONS = {
    "I": (parse_bc_int, stream_bc_int),
    "S": (parse_bc_string, stream_bc_string),
    "h": (lambda f: struct.unpack("!H", f.read(2))[0], lambda f, v: f.write(struct.pack("!H", v))),
    "L": (lambda f: struct.unpack("<L", f.read(4))[0], lambda f, v: f.write(struct.pack("<L", v))),
    "Q": (lambda f: struct.unpack("<Q", f.read(8))[0], lambda f, v: f.write(struct.pack("<Q", v))),
    "#": (lambda f: f.read(32), lambda f, v: f.write(v[:32])),
    "@": (lambda f: f.read(16), lambda f, v: f.write(v[:16])),
    "b": (lambda f: struct.unpack("?", f.read(1))[0], lambda f, b: f.write(struct.pack("?", b))),
}

BITCOIN_STREAMER = Streamer()
BITCOIN_STREAMER.register_array_count_parse(parse_bc_int)
BITCOIN_STREAMER.register_functions(STREAMER_FUNCTIONS.items())

parse_struct = BITCOIN_STREAMER.parse_struct
parse_as_dict = BITCOIN_STREAMER.parse_as_dict
stream_struct = BITCOIN_STREAMER.stream_struct
pack_struct = BITCOIN_STREAMER.pack_struct



# -*- coding: utf-8 -*-
"""
Parse and stream Bitcoin blocks as either Block or BlockHeader structures.


The MIT License (MIT)

Copyright (c) 2013 by Richard Kiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import struct

import io

from .encoding import double_sha256
from .merkle import merkle
from .serialize.bitcoin_streamer import parse_struct, stream_struct
from .serialize import b2h, b2h_rev

from .tx import Tx


class BadMerkleRootError(Exception):
    pass


def difficulty_max_mask_for_bits(bits):
    prefix = bits >> 24
    mask = (bits & 0x7ffff) << (8 * (prefix - 3))
    return mask


class BlockHeader(object):
    """A BlockHeader is a block with the transaction data removed. With a
    complete Merkle tree database, it can be reconstructed from the
    merkle_root."""

    @classmethod
    def parse(self, f):
        """Parse the BlockHeader from the file-like object in the standard way
        that blocks are sent in the network (well, except we ignore the
        transaction information)."""
        (version, previous_block_hash, merkle_root,
            timestamp, difficulty, nonce) = struct.unpack("<L32s32sLLL", f.read(4+32+32+4*3))
        return self(version, previous_block_hash, merkle_root, timestamp, difficulty, nonce)

    def __init__(self, version, previous_block_hash, merkle_root, timestamp, difficulty, nonce):
        self.version = version
        self.previous_block_hash = previous_block_hash
        self.merkle_root = merkle_root
        self.timestamp = timestamp
        self.difficulty = difficulty
        self.nonce = nonce

    def hash(self):
        """Calculate the hash for the block header. Note that this has the bytes
        in the opposite order from how the header is usually displayed (so the
        long string of 00 bytes is at the end, not the beginning)."""
        if not hasattr(self, "__hash"):
            s = io.BytesIO()
            self.stream_header(s)
            self.__hash = double_sha256(s.getvalue())
        return self.__hash

    def stream_header(self, f):
        """Stream the block header in the standard way to the file-like object f."""
        stream_struct("L##LLL", f, self.version, self.previous_block_hash,
                      self.merkle_root, self.timestamp, self.difficulty, self.nonce)

    def stream(self, f):
        """Stream the block header in the standard way to the file-like object f.
        The Block subclass also includes the transactions."""
        self.stream_header(f)

    def id(self):
        """Returns the hash of the block displayed with the bytes in the order
        they are usually displayed in."""
        return b2h_rev(self.hash())

    def previous_block_id(self):
        """Returns the hash of the previous block, with the bytes in the order
        they are usually displayed in."""
        return b2h_rev(self.previous_block_hash)

    def __str__(self):
        return "BlockHeader [%s] (previous %s)" % (self.id(), self.previous_block_id())

    def __repr__(self):
        return "BlockHeader [%s] (previous %s)" % (self.id(), self.previous_block_id())


class Block(BlockHeader):
    """A Block is an element of the Bitcoin chain. Generating a block
    yields a reward!"""

    @classmethod
    def parse(self, f):
        """Parse the Block from the file-like object in the standard way
        that blocks are sent in the network."""
        (version, previous_block_hash, merkle_root, timestamp,
            difficulty, nonce, count) = parse_struct("L##LLLI", f)
        txs = []
        for i in range(count):
            txs.append(Tx.parse(f))
        return self(version, previous_block_hash, merkle_root, timestamp, difficulty, nonce, txs)

    def __init__(self, version, previous_block_hash, merkle_root, timestamp, difficulty, nonce, txs):
        self.version = version
        self.previous_block_hash = previous_block_hash
        self.merkle_root = merkle_root
        self.timestamp = timestamp
        self.difficulty = difficulty
        self.nonce = nonce
        self.txs = txs

    def stream(self, f):
        """Stream the block in the standard way to the file-like object f."""
        stream_struct("L##LLLI", f, self.version, self.previous_block_hash,
                      self.merkle_root, self.timestamp, self.difficulty, self.nonce, len(self.txs))
        for t in self.txs:
            t.stream(f)

    def check_merkle_hash(self):
        """Raise a BadMerkleRootError if the Merkle hash of the
        transactions does not match the Merkle hash included in the block."""
        calculated_hash = merkle([tx.hash() for tx in self.txs], double_sha256)
        if calculated_hash != self.merkle_root:
            raise BadMerkleRootError(
                "calculated %s but block contains %s" % (b2h(calculated_hash), b2h(self.merkle_root)))

    def __str__(self):
        return "Block [%s] (previous %s) [tx count: %d]" % (
            self.id(), self.previous_block_id(), len(self.txs))

    def __repr__(self):
        return "Block [%s] (previous %s) [tx count: %d] %s" % (
            self.id(), self.previous_block_id(), len(self.txs), self.txs)



# -*- coding: utf-8 -*-
"""
Parse, stream, create, sign and verify Bitcoin transactions as Tx structures.


The MIT License (MIT)

Copyright (c) 2013 by Richard Kiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

from ... import ecdsa
from ...encoding import sec_to_public_pair, EncodingError

from . import der
from . import opcodes
from . import ScriptError

from .microcode import MICROCODE_LOOKUP, VCH_TRUE, VCH_FALSE, make_bool
from .tools import get_opcode

VERIFY_OPS = frozenset((opcodes.OPCODE_TO_INT[s] for s in "OP_NUMEQUALVERIFY OP_EQUALVERIFY OP_CHECKSIGVERIFY OP_VERIFY OP_CHECKMULTISIGVERIFY".split()))

INVALID_OPCODE_VALUES = frozenset((opcodes.OPCODE_TO_INT[s] for s in "OP_CAT OP_SUBSTR OP_LEFT OP_RIGHT OP_INVERT OP_AND OP_OR OP_XOR OP_2MUL OP_2DIV OP_MUL OP_DIV OP_MOD OP_LSHIFT OP_RSHIFT".split()))

def parse_signature_blob(sig_blob):
    sig_pair = der.sigdecode_der(sig_blob[:-1])
    signature_type = ord(sig_blob[-1:])
    return sig_pair, signature_type


def eval_script(script, signature_for_hash_type_f, expected_hash_type=None, stack=[]):
    altstack = []
    if len(script) > 10000:
        return False

    pc = 0
    begin_code_hash = pc
    if_condition = None # or True or False
    op_count = 0
    # TODO: set op_count

    try:
        while pc < len(script):
            opcode, data, pc = get_opcode(script, pc)
            if len(data) > 0:
                stack.append(data)
                continue

            # deal with if_condition first

            if if_condition is not None:
                # TODO: fix IF (which doesn't properly nest)
                if opcode == opcodes.OP_ELSE:
                    if_condition = not if_condition
                    continue
                if opcode == opcodes.OP_ENDIF:
                    if_condition = None
                    continue
                if not if_condition:
                    continue
            if opcode in (opcodes.OP_IF, opcodes.OP_NOTIF):
                if_condition = (stack.pop() == VCH_TRUE)
                continue

            if opcode == opcodes.OP_CODESEPARATOR:
                begin_code_hash = pc - 1
                continue

            if opcode in INVALID_OPCODE_VALUES:
                raise ScriptError("invalid opcode %s at %d" % (opcodes.INT_TO_OPCODE[opcode], pc-1))

            if opcode in MICROCODE_LOOKUP:
                MICROCODE_LOOKUP[opcode](stack)
                if opcode in VERIFY_OPS:
                    v = stack.pop()
                    if v != VCH_TRUE:
                        raise ScriptError("VERIFY failed at %d" % (pc-1))
                continue

            if opcode == opcodes.OP_TOALTSTACK:
                altstack.append(stack.pop())
                continue

            if opcode == opcodes.OP_FROMALTSTACK:
                stack.append(altstack.pop())
                continue

            if opcode >= opcodes.OP_1NEGATE and opcode <= opcodes.OP_16:
                stack.append(opcode + 1 - opcodes.OP_1)
                continue

            if opcode in (opcodes.OP_ELSE, opcodes.OP_ENDIF):
                raise ScriptError("%s without OP_IF" % opcodes.INT_TO_OPCODE[opcode])

            if opcode in (opcodes.OP_CHECKSIG, opcodes.OP_CHECKSIGVERIFY):
                public_pair = sec_to_public_pair(stack.pop())
                sig_pair, signature_type = parse_signature_blob(stack.pop())
                if expected_hash_type not in (None, signature_type):
                    raise ScriptError("wrong hash type")
                signature_hash = signature_for_hash_type_f(signature_type)
                if ecdsa.verify(ecdsa.generator_secp256k1, public_pair, signature_hash, sig_pair):
                    stack.append(VCH_TRUE)
                else:
                    stack.append(VCH_FALSE)
                if opcode == opcodes.OP_CHECKSIGVERIFY:
                    if stack.pop() != VCH_TRUE:
                        raise ScriptError("VERIFY failed at %d" % pc-1)
                continue

            if opcode == opcodes.OP_CHECKMULTISIG:
                key_count = stack.pop()
                public_pairs = []
                for i in range(key_count):
                    the_sec = stack.pop()
                    try:
                        public_pairs.append(sec_to_public_pair(the_sec))
                    except EncodingError:
                        # we must ignore badly encoded public pairs
                        # the transaction 70c4e749f2b8b907875d1483ae43e8a6790b0c8397bbb33682e3602617f9a77a
                        # is in a block and requires this hack
                        pass

                signature_count = stack.pop()
                sig_blobs = []
                for i in range(signature_count):
                    sig_blobs.append(stack.pop())

                should_be_zero_bug = stack.pop()

                sig_ok = VCH_TRUE
                for sig_blob in sig_blobs:
                    sig_pair, signature_type = parse_signature_blob(sig_blob)
                    signature_hash = signature_for_hash_type_f(signature_type)

                    ppp = ecdsa.possible_public_pairs_for_signature(
                        ecdsa.generator_secp256k1, signature_hash, sig_pair)

                    ppp.intersection_update(public_pairs)
                    if len(ppp) == 0:
                        sig_ok = VCH_FALSE
                        break

                    matching_pair = ppp.pop()
                    idx = public_pairs.index(matching_pair)
                    public_pairs = public_pairs[:idx] + public_pairs[idx+1:]

                stack.append(sig_ok)
                continue

            # BRAIN DAMAGE -- does it always get down here for each verify op? I think not
            if opcode in VERIFY_OPS:
                v = stack.pop()
                if v != VCH_TRUE:
                    raise ScriptError("VERIFY failed at %d" % pc-1)

            logging.error("can't execute opcode %s", opcode)

    except Exception as ex:
        logging.exception("script failed")

    return len(stack) != 0

def verify_script(script_signature, script_public_key, signature_for_hash_type_f, expected_hash_type=None):
    stack = []

    is_p2h = (len(script_public_key) == 23 and script_public_key[0] == opcodes.OP_HASH160
                and script_public_key[-1] == opcodes.OP_EQUAL)

    if not eval_script(script_signature, signature_for_hash_type_f, expected_hash_type, stack):
        logging.debug("script_signature did not evaluate")
        return False

    if is_p2h:
        signatures, alt_script_public_key = stack[:-1], stack[-1]
        from pycoin.tx.script import tools
        from pycoin import serialize
        def sub(x):
            if x == '00':
                return '0'
            return x
        s1 = [sub(serialize.b2h(s)) for s in signatures]
        alt_script_signature = tools.compile(" ".join(s1))

    if not eval_script(script_public_key, signature_for_hash_type_f, expected_hash_type, stack):
        logging.debug("script_public_key did not evaluate")
        return False

    if is_p2h and stack[-1] == VCH_TRUE:
        return verify_script(alt_script_signature, alt_script_public_key,
                             signature_for_hash_type_f, expected_hash_type=expected_hash_type)

    return stack[-1] == VCH_TRUE



"""
Provide the following functions:

bytes_to_ints(bytes):
    yield an iterator of ints. Designed to deal with how
    Python 2 treats bytes[0] as a string while
    Python 3 treats bytes[0] as an int.

to_bytes(v, length, byteorder):
    convert integer v into a bytes object

from_bytes(bytes, byteorder, *, signed=False):
    convert the bytes object into an integer

The last two functions are designed to mimic the methods of the same
name that exist on int in Python 3 only. For Python 3, use
those implementations.
"""

bytes_to_ints = (lambda x: (ord(c) for c in x)) if bytes == str else lambda x: x

if hasattr(int, "to_bytes"):
    to_bytes = lambda v, length, byteorder="big": v.to_bytes(length, byteorder=byteorder)
    from_bytes = lambda bytes, byteorder="big", signed=False: int.from_bytes(bytes, byteorder="big", signed=signed)
else:
    def to_bytes(v, length, byteorder="big"):
        l = bytearray()
        for i in range(length):
            mod = v & 0xff
            v >>= 8
            l.append(mod)
        if byteorder == "big":
            l.reverse()
        return bytes(l)

    def from_bytes(bytes, byteorder="big", signed=False):
        if byteorder != "big":
            bytes = reversed(bytes)
        v = 0
        for c in bytes_to_ints(bytes):
            v <<= 8
            v += c
        if signed and bytes[0] & 0x80:
            v = v - (1<<(8*len(bytes)))
        return v
